@@ -1,6 +1,6 @@
-
-use lib "/usr/local/libdata/perl5/site_perl/CIL"; # The libraries are in the lib directory
+use lib "/usr/local/libdata/perl5/site_perl/CIL";
 $::archos    = "libexec/cil";
 $::cc        = "/usr/bin/gcc";
 $::cilhome   = "/usr/local";
-$::default_mode = "GNUCC";
\ No newline at end of file
+$::default_mode = "GNUCC";
+
